@@ -45,6 +45,42 @@ Packet_courier * courierp = 0;		// local vars needed by add_member
 unsigned long long partial_data_size = 0;	// size of current block
 
 
+class Slot_tally
+  {
+  const int num_slots;				// total slots
+  int num_free;					// remaining free slots
+  pthread_mutex_t mutex;
+  pthread_cond_t slot_av;			// slot available
+
+  Slot_tally( const Slot_tally & );		// declared as private
+  void operator=( const Slot_tally & );		// declared as private
+
+public:
+  explicit Slot_tally( const int slots )
+    : num_slots( slots ), num_free( slots )
+    { xinit_mutex( &mutex ); xinit_cond( &slot_av ); }
+
+  ~Slot_tally() { xdestroy_cond( &slot_av ); xdestroy_mutex( &mutex ); }
+
+  bool all_free() { return ( num_free == num_slots ); }
+
+  void get_slot()				// wait for a free slot
+    {
+    xlock( &mutex );
+    while( num_free <= 0 ) xwait( &slot_av, &mutex );
+    --num_free;
+    xunlock( &mutex );
+    }
+
+  void leave_slot()				// return a slot to the tally
+    {
+    xlock( &mutex );
+    if( ++num_free == 1 ) xsignal( &slot_av );	// num_free was 0
+    xunlock( &mutex );
+    }
+  };
+
+
 struct Ipacket			// filename, file size and headers
   {
   const unsigned long long file_size;
@@ -332,7 +368,11 @@ void loop_encode( const uint8_t * const ibuf, const int isize,
       courier.collect_packet( new Opacket( obuf, opos ), worker_id );
       opos = 0; obuf = new( std::nothrow ) uint8_t[max_packet_size];
       if( !obuf ) { show_error( mem_msg2 ); cleanup_and_fail(); }
-      if( LZ_compress_finished( encoder ) == 1 ) break;
+      if( LZ_compress_finished( encoder ) == 1 )
+        {
+        if( LZ_compress_restart_member( encoder, LLONG_MAX ) >= 0 ) break;
+        show_error( "LZ_compress_restart_member failed." ); cleanup_and_fail();
+        }
       }
     }
   if( ipos > isize ) internal_error( "ipacket size exceeded in worker." );
@@ -365,27 +405,27 @@ extern "C" void * cworker( void * arg )
   if( !rbuf.size() ) { show_error( mem_msg2 ); cleanup_and_fail(); }
 
   int opos = 0;
+  bool flushed = true;		// avoid producing empty lzip members
   while( true )
     {
     const Ipacket * const ipacket = courier.distribute_packet( worker_id );
     if( !ipacket ) break;		// no more packets to process
-    if( ipacket->filename.empty() )	// end of group, flush encoder
+    if( ipacket->filename.empty() )	// end of group
       {
-      if( !encoder ) { delete ipacket; continue; }	// nothing to flush
-      loop_encode( 0, 0, data, opos, courier, encoder, worker_id, true );
+      if( !flushed )			// this lzip member is not empty
+        loop_encode( 0, 0, data, opos, courier, encoder, worker_id, true );
       courier.collect_packet( new Opacket, worker_id );	// end of member token
-      if( LZ_compress_restart_member( encoder, LLONG_MAX ) < 0 )
-        { show_error( "LZ_compress_restart_member failed." ); cleanup_and_fail(); }
-      delete ipacket; continue;
+      flushed = true; delete ipacket; continue;
       }
 
     const int infd =
       ipacket->file_size ? open_instream( ipacket->filename.c_str() ) : -1;
-    if( ipacket->file_size && infd < 0 )
+    if( ipacket->file_size && infd < 0 )	// can't read file data
       { delete[] ipacket->header; delete ipacket->extended; delete ipacket;
-        set_error_status( 1 ); continue; }
+        set_error_status( 1 ); continue; }	// skip file
 
-    if( !encoder )
+    flushed = false;
+    if( !encoder )		// init encoder just before using it
       {
       data = new( std::nothrow ) uint8_t[max_packet_size];
       encoder = LZ_compress_open( dictionary_size, match_len_limit, LLONG_MAX );
@@ -458,36 +498,35 @@ extern "C" void * cworker( void * arg )
 
 /* Get from courier the processed and sorted packets, and write
    their contents to the output archive. */
-bool muxer( Packet_courier & courier, const char * const archive_name,
-            const int outfd )
+void muxer( Packet_courier & courier, const int outfd )
   {
   while( true )
     {
     const Opacket * const opacket = courier.deliver_packet();
     if( !opacket ) break;	// queue is empty. all workers exited
 
-    const int wr = writeblock( outfd, opacket->data, opacket->size );
-    if( wr != opacket->size )
-      { show_file_error( archive_name, "Write error", errno ); return false; }
+    if( !writeblock_wrapper( outfd, opacket->data, opacket->size ) )
+      cleanup_and_fail();
     delete[] opacket->data;
     delete opacket;
     }
-  return true;
   }
 
 } // end namespace
 
 
 // init the courier, then start the grouper and the workers and call the muxer
-int encode_lz( const char * const archive_name, const Arg_parser & parser,
-               const int dictionary_size, const int match_len_limit,
-               const int num_workers, const int outfd, const int debug_level )
+int encode_lz( const Arg_parser & parser, const int dictionary_size,
+               const int match_len_limit, const int num_workers,
+               const int outfd, const int debug_level )
   {
   const int in_slots = 65536;		// max small files (<=512B) in 64 MiB
   const int total_in_slots = ( INT_MAX / num_workers >= in_slots ) ?
                              num_workers * in_slots : INT_MAX;
   const int out_slots = 64;
 
+  /* If an error happens after any threads have been started, exit must be
+     called before courier goes out of scope. */
   Packet_courier courier( num_workers, total_in_slots, out_slots );
   courierp = &courier;			// needed by add_member
 
@@ -498,11 +537,12 @@ int encode_lz( const char * const archive_name, const Arg_parser & parser,
   pthread_t grouper_thread;
   int errcode = pthread_create( &grouper_thread, 0, grouper, &grouper_arg );
   if( errcode )
-    { show_error( "Can't create grouper thread", errcode ); return 1; }
+    { show_error( "Can't create grouper thread", errcode ); cleanup_and_fail(); }
 
   Worker_arg * worker_args = new( std::nothrow ) Worker_arg[num_workers];
   pthread_t * worker_threads = new( std::nothrow ) pthread_t[num_workers];
-  if( !worker_args || !worker_threads ) { show_error( mem_msg ); return 1; }
+  if( !worker_args || !worker_threads )
+    { show_error( mem_msg ); cleanup_and_fail(); }
   for( int i = 0; i < num_workers; ++i )
     {
     worker_args[i].courier = &courier;
@@ -511,23 +551,23 @@ int encode_lz( const char * const archive_name, const Arg_parser & parser,
     worker_args[i].worker_id = i;
     errcode = pthread_create( &worker_threads[i], 0, cworker, &worker_args[i] );
     if( errcode )
-      { show_error( "Can't create worker threads", errcode ); return 1; }
+      { show_error( "Can't create worker threads", errcode ); cleanup_and_fail(); }
     }
 
-  if( !muxer( courier, archive_name, outfd ) ) return 1;
+  muxer( courier, outfd );
 
   for( int i = num_workers - 1; i >= 0; --i )
     {
     errcode = pthread_join( worker_threads[i], 0 );
     if( errcode )
-      { show_error( "Can't join worker threads", errcode ); return 1; }
+      { show_error( "Can't join worker threads", errcode ); cleanup_and_fail(); }
     }
   delete[] worker_threads;
   delete[] worker_args;
 
   errcode = pthread_join( grouper_thread, 0 );
   if( errcode )
-    { show_error( "Can't join grouper thread", errcode ); return 1; }
+    { show_error( "Can't join grouper thread", errcode ); cleanup_and_fail(); }
 
   // write End-Of-Archive records
   int retval = 0;
@@ -537,9 +577,7 @@ int encode_lz( const char * const archive_name, const Arg_parser & parser,
     0xA3, 0xB7, 0x80, 0x0C, 0x82, 0xDB, 0xFF, 0xFF, 0x9F, 0xF0, 0x00, 0x00,
     0x2E, 0xAF, 0xB5, 0xEF, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
     0x2C, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
-  if( writeblock( outfd, eof_member, eof_member_size ) != eof_member_size )
-    { show_error( "Error writing end-of-archive blocks", errno );
-      retval = 1; }
+  if( !writeblock_wrapper( outfd, eof_member, eof_member_size ) ) retval = 1;
 
   if( close( outfd ) != 0 && !retval )
     { show_error( "Error closing archive", errno ); retval = 1; }
