@@ -19,10 +19,12 @@
 
 #include <algorithm>
 #include <cerrno>
+#include <climits>
 #include <cstdio>
 #include <cstring>
 #include <string>
 #include <vector>
+#include <pthread.h>
 #include <stdint.h>
 #include <unistd.h>
 
@@ -30,8 +32,6 @@
 #include "tarlz.h"
 
 
-namespace {
-
 int seek_read( const int fd, uint8_t * const buf, const int size,
                const long long pos )
   {
@@ -40,6 +40,8 @@ int seek_read( const int fd, uint8_t * const buf, const int size,
   return 0;
   }
 
+namespace {
+
 const char * bad_version( const unsigned version )
   {
   static char buf[80];
