@@ -68,6 +68,7 @@ lzlib_1_11() { [ ${lwarn} = 0 ] &&
 # t155.tar[.lz]:     directory + links + file + eof, all with 155 char names
 # tar_in_tlz1.tar.lz 2 members (test.txt.tar test3.tar) 3 lzip members
 # tar_in_tlz2.tar.lz 2 members (test.txt.tar test3.tar) 5 lzip members
+# ts_in_link.tar.lz: 4 symbolic links (link[1-4]) to / /dir/ dir/ dir(107/)
 # test_bad1.tar.lz:  truncated at offset 6000 (of 7495)
 # test_bad2.tar.lz:  byte at offset 6000 changed from 0x56 to 0x46
 # test3.tar:         3 members (foo bar baz) + 2 zeroed 512-byte blocks
@@ -89,9 +90,12 @@ lzlib_1_11() { [ ${lwarn} = 0 ] &&
 # test3_eof1.tar.lz: test3.tar.lz without eof blocks
 # test3_eof2.tar.lz: test3.tar.lz with only one eof block
 # test3_eof3.tar.lz: test3.tar.lz with one zeroed block between foo and bar
+# test3_em?.tar.lz:  test3.tar.lz with one empty lzip member at each position
+# test3_em6.tar.lz:  test3.tar.lz preceded by four empty lzip members
 # tlz_in_tar1.tar:   1 member (test3.tar.lz) first magic damaged
 # tlz_in_tar2.tar:   2 members (foo test3.tar.lz) first magic damaged
 # ug32chars.tar.lz:  1 member (foo) with 32-character owner and group names
+# ug32767.tar.lz:    1 member (foo) with numerical-only owner and group
 
 printf "testing tarlz-%s..." "$2"
 
@@ -262,7 +266,8 @@ cmp cfoo foo || test_failed $LINENO
 [ ! -e bar ] || test_failed $LINENO
 [ ! -e baz ] || test_failed $LINENO
 rm -f foo bar baz || framework_failure
-#
+
+# test --list and --extract tar in tar.lz
 for i in "${tarint1_lz}" "${tarint2_lz}" ; do
 	for j in 0 2 6 ; do
 		"${TARLZ}" -tf "$i" --threads=$j > out$j ||
@@ -270,19 +275,44 @@ for i in "${tarint1_lz}" "${tarint2_lz}" ; do
 		"${TARLZ}" -tvf "$i" --threads=$j > outv$j ||
 			test_failed $LINENO "$i $j"
 	done
-	cmp out0 out2 || test_failed $LINENO
-	cmp out0 out6 || test_failed $LINENO
-	cmp out2 out6 || test_failed $LINENO
-	cmp outv0 outv2 || test_failed $LINENO
-	cmp outv0 outv6 || test_failed $LINENO
-	cmp outv2 outv6 || test_failed $LINENO
+	diff -u out0 out2 || test_failed $LINENO $i
+	diff -u out0 out6 || test_failed $LINENO $i
+	diff -u out2 out6 || test_failed $LINENO $i
+	diff -u outv0 outv2 || test_failed $LINENO $i
+	diff -u outv0 outv6 || test_failed $LINENO $i
+	diff -u outv2 outv6 || test_failed $LINENO $i
 	rm -f out0 out2 out6 outv0 outv2 outv6 || framework_failure
-	"${TARLZ}" -xf "$i" || test_failed $LINENO
-	cmp "${in_tar}" test.txt.tar || test_failed $LINENO
-	cmp "${test3}" test3.tar || test_failed $LINENO
+	"${TARLZ}" -xf "$i" || test_failed $LINENO $i
+	cmp "${in_tar}" test.txt.tar || test_failed $LINENO $i
+	cmp "${test3}" test3.tar || test_failed $LINENO $i
 	rm -f test.txt.tar test3.tar || framework_failure
 done
 
+# test --list and --extract with empty lzip members
+for i in 1 2 3 4 5 6 ; do
+	for j in 0 2 6 ; do
+		"${TARLZ}" -tf "${testdir}"/test3_em${i}.tar.lz --threads=$j \
+			> out$j || test_failed $LINENO "$i $j"
+		"${TARLZ}" -tvf "${testdir}"/test3_em${i}.tar.lz --threads=$j \
+			> outv$j || test_failed $LINENO "$i $j"
+	done
+	diff -u out0 out2 || test_failed $LINENO $i
+	diff -u out0 out6 || test_failed $LINENO $i
+	diff -u out2 out6 || test_failed $LINENO $i
+	diff -u outv0 outv2 || test_failed $LINENO $i
+	diff -u outv0 outv6 || test_failed $LINENO $i
+	diff -u outv2 outv6 || test_failed $LINENO $i
+	rm -f out0 out2 out6 outv0 outv2 outv6 || framework_failure
+	for j in 0 2 6 ; do
+		"${TARLZ}" -xf "${testdir}"/test3_em${i}.tar.lz --threads=$j ||
+			test_failed $LINENO "$i $j"
+		cmp cfoo foo || test_failed $LINENO "$i $j"
+		cmp cbar bar || test_failed $LINENO "$i $j"
+		cmp cbaz baz || test_failed $LINENO "$i $j"
+		rm -f foo bar baz || framework_failure
+	done
+done
+
 # test --concatenate
 cat "${in_tar_lz}" > out.tar.lz || framework_failure
 "${TARLZ}" -Af out.tar.lz "${test3_lz}" || test_failed $LINENO
@@ -421,10 +451,11 @@ cmp out.tar.lz aout.tar.lz || test_failed $LINENO
 rm -f out.tar.lz aout.tar.lz || framework_failure
 
 # append to solid archive
-"${TARLZ}" --solid -0 -cf out.tar.lz foo || test_failed $LINENO
+"${TARLZ}" --solid -q -0 -cf out.tar.lz "${in}" foo bar || test_failed $LINENO
+"${TARLZ}" -q -tf out.tar.lz || test_failed $LINENO	# compressed seekable
 cat out.tar.lz > aout.tar.lz || framework_failure
 for i in --asolid --bsolid --dsolid --solid -0 ; do
-	"${TARLZ}" $i -q -rf out.tar.lz bar baz
+	"${TARLZ}" $i -q -rf out.tar.lz baz
 	[ $? = 2 ] || test_failed $LINENO $i
 	cmp out.tar.lz aout.tar.lz || test_failed $LINENO $i
 done
@@ -445,6 +476,19 @@ for i in --asolid --bsolid --dsolid -0 ; do
 done
 rm -f foo bar baz || framework_failure
 
+# test --diff
+"${TARLZ}" -xf "${test3_lz}" || test_failed $LINENO
+"${TARLZ}" --uncompressed -cf out.tar foo || test_failed $LINENO
+"${TARLZ}" --uncompressed -cf aout.tar foo --anonymous || test_failed $LINENO
+if cmp out.tar aout.tar > /dev/null ; then
+	printf "\nwarning: --diff test can't be run as root."
+else
+	"${TARLZ}" -q -df "${test3_lz}"
+	[ $? = 1 ] || test_failed $LINENO
+	"${TARLZ}" -df "${test3_lz}" --ignore-ids || test_failed $LINENO
+fi
+rm -f out.tar aout.tar foo bar baz || framework_failure
+
 # test directories and links
 mkdir dir1 || framework_failure
 "${TARLZ}" -0 -cf out.tar.lz dir1 || test_failed $LINENO
@@ -475,12 +519,16 @@ if ln dummy_file dummy_link 2> /dev/null &&
 	ln dir1/dir2/dir3/in dir1/dir2/dir3/"${name_100}" || framework_failure
 	ln dir1/dir2/dir3/in "${path_100}" || framework_failure
 	ln dir1/dir2/dir3/in "${path_106}" || framework_failure
+	ln -s dir2/ dir1/dir2_link || framework_failure
 	ln -s in dir1/dir2/dir3/link || framework_failure
 	ln -s "${name_100}" dir1/dir2/dir3/link_100 || framework_failure
 	"${TARLZ}" -0 -cf out.tar.lz dir1 || test_failed $LINENO
+	"${TARLZ}" -df out.tar.lz || test_failed $LINENO
 	rm -rf dir1 || framework_failure
 	"${TARLZ}" -xf out.tar.lz || test_failed $LINENO
+	"${TARLZ}" -df out.tar.lz || test_failed $LINENO
 	cmp "${in}" dir1/dir2/dir3/in || test_failed $LINENO
+	cmp "${in}" dir1/dir2_link/dir3/in || test_failed $LINENO
 	cmp "${in}" dir1/dir2/dir3/"${name_100}" || test_failed $LINENO
 	cmp "${in}" "${path_100}" || test_failed $LINENO
 	cmp "${in}" "${path_106}" || test_failed $LINENO
@@ -494,8 +542,12 @@ if ln dummy_file dummy_link 2> /dev/null &&
 	cmp "${in}" dir1/dir2/dir3/in || test_failed $LINENO
 	cmp "${in}" dir1/dir2/dir3/link || test_failed $LINENO
 	"${TARLZ}" -0 -q -c ../tmp/dir1 | "${TARLZ}" -x || test_failed $LINENO
-	diff -r tmp/dir1 dir1 || test_failed $LINENO
+	diff -ru tmp/dir1 dir1 || test_failed $LINENO
 	rm -rf tmp/dir1 dir1 || framework_failure
+	"${TARLZ}" -xf "${testdir}"/ts_in_link.tar.lz || test_failed $LINENO
+	"${TARLZ}" -df "${testdir}"/ts_in_link.tar.lz --ignore-ids ||
+		test_failed $LINENO
+	rm -f link1 link2 link3 link4 || framework_failure
 else
 	printf "\nwarning: skipping link test: 'ln' does not work on your system."
 fi
@@ -510,11 +562,11 @@ if [ "${ln_works}" = yes ] ; then
 	"${TARLZ}" -C dir1 -xf "${t155}" || test_failed $LINENO
 	mkdir dir2 || framework_failure
 	"${TARLZ}" -C dir2 -xf "${t155_lz}" || test_failed $LINENO
-	diff -r dir1 dir2 || test_failed $LINENO
+	diff -ru dir1 dir2 || test_failed $LINENO
 	"${TARLZ}" -cf out.tar.lz dir2 || test_failed $LINENO
 	rm -rf dir2 || framework_failure
 	"${TARLZ}" -xf out.tar.lz || test_failed $LINENO
-	diff -r dir1 dir2 || test_failed $LINENO
+	diff -ru dir1 dir2 || test_failed $LINENO
 	rmdir dir2 2> /dev/null && test_failed $LINENO
 	rmdir dir1 2> /dev/null && test_failed $LINENO
 	rm -rf out.tar.lz dir2 dir1 || framework_failure
@@ -532,6 +584,11 @@ fi
 "${TARLZ}" -xf "${testdir}"/ug32chars.tar.lz || test_failed $LINENO
 cmp cfoo foo || test_failed $LINENO
 rm -f foo || framework_failure
+"${TARLZ}" -tvf "${testdir}"/ug32767.tar.lz | grep -q -e 32767/32767 ||
+	test_failed $LINENO
+"${TARLZ}" -xf "${testdir}"/ug32767.tar.lz || test_failed $LINENO
+cmp cfoo foo || test_failed $LINENO
+rm -f foo || framework_failure
 
 printf "\ntesting bad input..."
 
